/*
	picthp
	
	Original work by Andrew Low et.al
	
	QNX4 port and modifications by W. M.Brown , Dallas, TX
	
	NOTE THIS WILL NOT PROCESS THE CURRENT OL Print Window/Picture
		generated files.
		
		 Use something like this to capture:
--------------------------------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <windows/Qwindows.h>

void main( )
{
    int i, sid, pid, wid;
    QW_RECT_AREA ar;
    FILE *fp;
	
	sid = GraphicsOpen( NULL );
	if( !sid ) {
		printf( "QNX Windows required\n" );
		exit( -1 );
		}

	ar.height = 2000;
	ar.width = 2000;
	ar.row = 0;
	ar.col = 0;
	fp = PrintOpen( "test.prt", (unsigned)ar.height, (unsigned)ar.width );
	PrintArea( fp, &ar, NULL );
	PrintClose( fp );
	
	GraphicsClose( sid );
} 
-------------------------------------------------------------------------------------
		
Makefile:

CFLAGS  = -ms -g -w4
LDFLAGS = -N 5000
	
picthp:print.c
    $(CC) $(CFLAGS) $(LDFLAGS) -DHP -o picthp print.c -b


Program flow: nothing in particular, then lots of magic will 
happen and it will work.. 
*/

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <windows.h>
#include <wind_struct.h>
#include <wind_file.h>
#include "qwprint.h"

#ifdef PR
#define PRINTSIZE	1920	/* total square dpi size 	*/
#endif
#ifdef HP
#define PRINTSIZE	2400	/* total square dpi size 	*/
#endif

/*	Global flags */
int		debug=FALSE,		/* dump debug info			*/
		mono=FALSE,			/* black and white output	*/
		lowres=FALSE,		/* lower resolution			*/
		noscale=FALSE;		/* no scaling ie: 1to1		*/
	
/*	Global variables */
FILE	*fp;
HDR		*element;
int		height,				/* picture height in tips 	*/
		width,				/* picture width in tips 	*/
		landscape;			/* picture format 			*/
float	scale;				/* picture scaling factor 	*/
char	linebuf[1024]; 		/* buffer for bitmap image	*/
char	printbuf[PRINTSIZE];/* buffer for print image	*/

#ifdef PR
char 	pr_out1[PRINTSIZE],	/* dither print data		*/
		pr_out2[PRINTSIZE],
		pr_out3[PRINTSIZE];
#endif
#ifdef HP
char	hp_out1[PRINTSIZE],	/* dither print data		*/
		hp_out2[PRINTSIZE],
		hp_out3[PRINTSIZE];
#endif

char	colour[3][16];		/* color mapping table		*/

/*	Dithering data */
static char dither[3][10]={
		{ 0, 0, 0, 0, 2, 5, 6, 5, 7, 7},
		{ 0, 2, 2, 5, 5, 2, 3, 7, 5, 7},
		{ 0, 0, 1, 2, 2, 5, 6, 5, 7, 7} };

main(argc,argv)
int argc;
char *argv[];
{
	int i;
	char *p;
	PICTHDR *picthdr;

/*	Get buffer for elements */
	if((element=(HDR *)malloc(8192))==NULL) {
		printf("Could not allocate element buffer\n");
		exit(-1);
		}
	
/*	Parse command line */
	if(argc<2)
		give_usage(argv[0]);
	i=2;
	while(i<argc) {
		p=argv[i];
		switch((*p<<8)|*(p+1)) {
			case '+d' :	debug=TRUE;			break;
			case '+m' :	mono=TRUE;			break;
			case '+1' :	lowres=TRUE;		break;
			case '-s' : noscale=TRUE;		break;
			default	  : give_usage(argv[0]);
			}
		i++;
		}

/*	Open input file */
	if((fp=fopen(argv[1],"r"))==NULL) {
		printf("ERROR: could not open %s\n",argv[1]);
		exit(-1);
		}

/*	read extended window file */

	read_extended_hdr(fp);
	read_element(fp);

	picthdr = (PICTHDR *)element;
	landscape = (picthdr->scale==0)?0:1;
	if(debug)
		printf("landscape:%d\n",landscape);

	if(landscape) {
		width = picthdr->height/11;
		height = picthdr->width/11;
		}
	else
		{
		height = picthdr->height/11;
		width = picthdr->width/11;
		}

/*	determine scaling */
	if(debug)
		printf("Width:%d Height:%d\n",width,height);
	if(mono) {
		if(width>height)
			scale=(float)(PRINTSIZE/(lowres?2:1))/(float)(width);
		else
			scale=(float)(PRINTSIZE/(lowres?2:1))/(float)(height);
		}
	else {
		if(width>height)
			scale=(float)(PRINTSIZE/(lowres?6:3))/(float)(width);
		else
			scale=(float)(PRINTSIZE/(lowres?6:3))/(float)(height);
		}
	if(noscale) {
		if(scale<1) 
			if(mono)
				width=(PRINTSIZE/(lowres?2:1));
			else
				width=(PRINTSIZE/(lowres?6:3));
		scale=1;
		}
	if(debug) printf("scale:%f",scale);
/*	process picture */	
	prheader();
    read_picture();

#ifdef PR
	printf("%c@",0x1b);
	printf("%c%c",0x0d,0x0a);
#endif
#ifdef HP
	printf("%c*rB",0x1b);
	printf("%cE",0x1b);
#endif
}

give_usage(p)                 
char *p;
{
	printf("Use: %s printfile [options] > output\n",p);
	printf("Options: +debug\n");
	printf("		 +mono output\n");
#ifdef PR
	printf("		 +120 dpi (default 240dpi)\n");
#endif
#ifdef HP
	printf("		 +150 dpi (default 300dpi)\n");
#endif
	printf("		 -scaling (ie: 1to1)\n\n");
	exit(1);
}

read_extended_hdr(fp)
FILE *fp;
{
	int i,n;
	PRINTHDR printhdr; 
#ifdef CII
	if (!fget (&printhdr,sizeof(PRINTHDR),fp)) {
#else
	if (!fread (&printhdr,sizeof(PRINTHDR),1,fp)) {
#endif
		printf("Could not extract print header\n");
		exit(-1);
		}

	if (printhdr.record_type != FILE_HEADER) {
		printf("ERROR: not an extended window file\n");
		exit(-1);
		}

	for (i=1;i<=printhdr.nbr_colors;i++) {
		if(debug) {
			printf("red:%f  ",printhdr.color[i-1].red / 9843.0 * 0.30);
			printf("green:%f  ",printhdr.color[i-1].green / 9843.0 * 0.59);
			printf("blue:%f  ",printhdr.color[i-1].blue / 9843.0 * 0.11 );
			printf("total:%f\n",printhdr.color[i-1].red / 9843.0 * 0.30 
		                  +printhdr.color[i-1].green / 9843.0 * 0.59
		           	      +printhdr.color[i-1].blue / 9843.0 * 0.11 );
			}
		n=(int)(9 * (1 - printhdr.color[i-1].red / 9843.0 * 0.30 
		                  -printhdr.color[i-1].green / 9843.0 * 0.59
		                  -printhdr.color[i-1].blue / 9843.0 * 0.11 ));
		if(mono) {
			colour[0][i-1]=(n>5)?15:0;
			if(debug)
				printf("n:%d\n",n);
			}
		else {
			colour[0][i-1]=dither[0][n];
			colour[1][i-1]=dither[1][n];
			colour[2][i-1]=dither[2][n];
			if(debug)
				printf("n:%d\n",n);
			}
		}
}

static read_element(fp)
	FILE *fp;
{
	char *elmt_offset;
	int len;

	elmt_offset = (char *)element;

	/* Read picture header */
#ifdef CII
	if (! (len = fget(element,sizeof(HDR),fp)))
#else
	 if (! (len = fread(element,1,sizeof(HDR),fp)))    
#endif
		return(0);

	/* End of file */
	if ( len < sizeof(HDR) )
		return (0);

	/* One of the print structures */
	if ((element->length < len) && element->length)
	{
		fseek(fp,-(long)(len - element->length),1);
		return(1);
	}
		
	elmt_offset += len;

	/* Read extra element info, if any */
	if (element->length - sizeof(HDR))
#ifdef CII
		if (!fget(elmt_offset,element->length-sizeof(HDR),fp))
#else
		 if (!fread(elmt_offset,element->length-sizeof(HDR),1,fp))  
#endif
			return(0);

	return(1);
}


read_picture()	/* This is a very sick and twisted function, but it works */
{
	int i,h;


	h=0;
	for(i=0;i<(scale>1)?((mono)?(PRINTSIZE/(lowres?2:1)):(PRINTSIZE/(lowres?6:3))):height;i++) {

/* read in next element */
		if(!read_element(fp)) 
			return;
		if (debug)
			printf("\nType: %c Tag: %s ",element->type,
			(char *)&element + sizeof(HDR));
		
		if ( (element->type == WIND_HEADER) || (element->type == PICT_HEADER) ||
			 (element->type == FILE_TRAILER) )
		   	return;
/* check if its really a pixmap */


/* scale the height */			
		if(scale>1) {
			/* decode the pixmap */
			if(landscape)
				decode_col((char *)element);
			else
				decode_row((char *)element);
			/* scale the width */
			scale_image();
			/*possibly repeat a line */
			while((int)(i/scale)<h) {
				if(debug)
					printf("print row:%d i:%d\n",h,i);
				output_line();
				i++;
				}
			i--;
			h++;
			}
		else {
			if((int)(scale*(float)i)==h) {
				/* decode the pixmap */
				if(landscape)
					decode_col((char *)element);
				else
					decode_row((char *)element);
				/* scale the width */
				scale_image();
				/*possibly drop a line */
				if(debug)
					printf("Print row:%d i:%d\n",h,i);
				output_line();
				h++;
				}
			}
		}
/* flush buffer */
	memset(printbuf,0,PRINTSIZE);
	while(output_line()!=0);
}



int decode_row(p)
char *p;
{
	HDR *header=(HDR *)p;
	IMAGE_DEF *image;
	char *data;
	int i, mask, bpl, n;

	image=(IMAGE_DEF *)(header+1);
	image=(IMAGE_DEF *)((char *)image+header->tag[0]+2);
	data=(char *)(image+1);

	if(image->type==256) {	/* vga card */
		bpl=image->bytes_per_line;
		for(i=0;i<image->width;i++) {
			n=i/8;
			mask=128>>(i&7);
			linebuf[i]=(char)((data[i/8]&mask)?1:0)+
					((data[n+bpl]&mask)?2:0)+
					((data[n+bpl*2]&mask)?4:0)+
					((data[n+bpl*3]&mask)?8:0);
			}
		}
	else 					/* 8514 card */
		memcpy(linebuf,data,image->width);

	return(image->width);
}


decode_col(p)
char *p;
{
	HDR *header=(HDR *)p;
	IMAGE_DEF *image;
	char *data, swap;
	int i, n /*, mask*/;

	image=(IMAGE_DEF *)(header+1);
	image=(IMAGE_DEF *)((char *)image+header->tag[0]+2);
	data=(char *)(image+1);

	if(image->type==256)	/* vga card */
		for(i=0;i<image->height;i++) {
			n=i*4;
			linebuf[i]=(char)((data[n]&128)?1:0)+
					((data[n+1]&128)?2:0)+
					((data[n+2]&128)?4:0)+
					((data[n+3]&128)?8:0);
			}
	else 					/* 8514 card */
		memcpy(linebuf,data,image->height);

/* reverse data */
	for(i=0;i<image->height/2;i++) {
		swap=linebuf[image->height-i-1];
		linebuf[image->height-i-1]=linebuf[i];
		linebuf[i]=swap;
		}
		
}

scale_image()	/*	scale 	*/
{
	int i;
    
	/* scale */
	for(i=0;i<width*scale;i++) 
		printbuf[i]=linebuf[(int)(i/scale)];

}

dither_image(bit)	/* dither the image */
int bit;
{
	int i, a, b;

#ifdef PR
	for(i=0;i<=width*scale;i++) {

		a=i*3;
		b=printbuf[i]&0x0f;
		pr_out1[a]|=(char)((colour[0][b]>>2&0x01)<<bit);
		pr_out2[a]|=(char)((colour[1][b]>>2&0x01)<<bit);
		pr_out3[a]|=(char)((colour[2][b]>>2&0x01)<<bit);
       
		pr_out1[a+1]|=(char)((colour[0][b]>>1&0x01)<<bit);
		pr_out2[a+1]|=(char)((colour[1][b]>>1&0x01)<<bit);
		pr_out3[a+1]|=(char)((colour[2][b]>>1&0x01)<<bit);
       	
		pr_out1[a+2]|=(char)((colour[0][b]&0x01)<<bit);
		pr_out2[a+2]|=(char)((colour[1][b]&0x01)<<bit);
		pr_out3[a+2]|=(char)((colour[2][b]&0x01)<<bit);
       	
		}
#endif
#ifdef HP
	bit = bit;
	for(i=0;i<=(width*scale/8);i++) {

		a=i*3;
		b=printbuf[i*8]&0x0f;
		hp_out1[a]|=(char)(colour[0][b]<<5);
		hp_out2[a]|=(char)(colour[1][b]<<5);
		hp_out3[a]|=(char)(colour[2][b]<<5);

		b=printbuf[i*8+1]&0x0f;
		hp_out1[a]|=(char)(colour[0][b]<<2);
		hp_out2[a]|=(char)(colour[1][b]<<2);
		hp_out3[a]|=(char)(colour[2][b]<<2);

		b=printbuf[i*8+2]&0x0f;
		hp_out1[a]|=(char)(colour[0][b]>>1);
		hp_out2[a]|=(char)(colour[1][b]>>1);
		hp_out3[a]|=(char)(colour[2][b]>>1);

		b=printbuf[i*8+2]&0x0f;
		hp_out1[a+1]|=(char)(colour[0][b]<<7);
		hp_out2[a+1]|=(char)(colour[1][b]<<7);
		hp_out3[a+1]|=(char)(colour[2][b]<<7);
	
		b=printbuf[i*8+3]&0x0f;
		hp_out1[a+1]|=(char)(colour[0][b]<<4);
		hp_out2[a+1]|=(char)(colour[1][b]<<4);
		hp_out3[a+1]|=(char)(colour[2][b]<<4);

		b=printbuf[i*8+4]&0x0f;
		hp_out1[a+1]|=(char)(colour[0][b]<<1);
		hp_out2[a+1]|=(char)(colour[1][b]<<1);
		hp_out3[a+1]|=(char)(colour[2][b]<<1);

		b=printbuf[i*8+5]&0x0f;
		hp_out1[a+1]|=(char)(colour[0][b]>>2);
		hp_out2[a+1]|=(char)(colour[1][b]>>2);
		hp_out3[a+1]|=(char)(colour[2][b]>>2);

		b=printbuf[i*8+5]&0x0f;
		hp_out1[a+2]|=(char)(colour[0][b]<<6);
		hp_out2[a+2]|=(char)(colour[1][b]<<6);
		hp_out3[a+2]|=(char)(colour[2][b]<<6);

		b=printbuf[i*8+6]&0x0f;
		hp_out1[a+2]|=(char)(colour[0][b]<<3);
		hp_out2[a+2]|=(char)(colour[1][b]<<3);
		hp_out3[a+2]|=(char)(colour[2][b]<<3);

		b=printbuf[i*8+7]&0x0f;
		hp_out1[a+2]|=(char)(colour[0][b]);
		hp_out2[a+2]|=(char)(colour[1][b]);
		hp_out3[a+2]|=(char)(colour[2][b]);

		}
/* trim excess information off print */
	hp_out1[(int)(width*scale*3)]=hp_out1[(int)(width*scale*3)]&~0xff>>(int)(width*scale*3)%8;
	hp_out2[(int)(width*scale*3)]=hp_out2[(int)(width*scale*3)]&~0xff>>(int)(width*scale*3)%8;
	hp_out2[(int)(width*scale*3)]=hp_out2[(int)(width*scale*3)]&~0xff>>(int)(width*scale*3)%8;
	
#endif
}

mono_image(bit)		/* convert the image to black and white */
int bit;
{
#ifdef PR
	int i, b;

	b=bit/3;
	switch(bit%3) {
		case 2:
			for(i=0;i<=width*scale;i++) {
				pr_out1[i]|=(char)((colour[0][printbuf[i]&0x0f]>>2&0x01)<<b);
				}
			break;
		case 1:
			for(i=0;i<=width*scale;i++) {
				pr_out2[i]|=(char)((colour[0][printbuf[i]&0x0f]>>2&0x01)<<b);
				}
			break;
		case 0:
			for(i=0;i<=width*scale;i++) {
				pr_out3[i]|=(char)((colour[0][printbuf[i]&0x0f]>>2&0x01)<<b);
				}
			break;
		}
#endif
#ifdef HP
	int i;

	switch(bit) {
		case 0:
			for(i=0;i<=width*scale;i++) {
				hp_out1[i/8]|=(char)((colour[0][printbuf[i]&0x0f])&(0x80>>(i&3)));
				}
			break;
		case 1:
			for(i=0;i<=width*scale;i++) {
				hp_out2[i/8]|=(char)((colour[0][printbuf[i]&0x0f])&(0x80>>(i&3)));
				}
			break;
		case 2:
			for(i=0;i<=width*scale;i++) {
				hp_out3[i/8]|=(char)((colour[0][printbuf[i]&0x0f])&(0x80>>(i&3)));
				}
			break;
		}
#endif

}

prheader()	/* print header */
{
#ifdef PR
	printf("%c@",0x1b);				/* reset printer 	*/
	if(lowres)
		printf("%c3%c",0x1b,0x02);	/* 2/216 linefeeds 	*/
	else
		printf("%c3%c",0x1b,0x01);	/* 1/216 linefeeds	*/
#endif
#ifdef HP
	printf("%cE",0x1b);   				/* Reset Printer 		*/
	printf("%c*rB",0x1b); 				/* Text mode 			*/
	printf("%c&l0O",0x1b);				/* Portrait mode 		*/
	if(lowres)
		printf("%c*t150R",0x1b);		/* 150 dpi resolution	*/
	else
		printf("%c*t300R",0x1b); 		/* 300 dpi resolution 	*/
	printf("%c*p+0X%c*p+0Y",0x1b,0x1b); /* where on the paper 	*/
	printf("%c*r1A",0x1b);  			/* raster mode 			*/
#endif	
}

print_line()	/* print the line stored in the printbuf */
{

#ifdef PR
	int i,n,lo,hi;
	char c;

	n=(int)(width*scale)*3;
	if(debug) printf("width*scale:%d",n);
	lo=n&0x00ff;
	hi=n>>8;

	if(lowres)
		printf("%cY%c%c",0x1b,lo,hi);
	else
		printf("%cZ%c%c",0x1b,lo,hi);	
	fput(pr_out1,n,stdout);	
	printf("%c%c",0x0d,0x0a);
	if(lowres)
		printf("%cY%c%c",0x1b,lo,hi);
	else
		printf("%cZ%c%c",0x1b,lo,hi);	
	fput(pr_out2,n,stdout);	
	printf("%c%c",0x0d,0x0a);
	if(lowres)
		printf("%cY%c%c",0x1b,lo,hi);
	else
		printf("%cZ%c%c",0x1b,lo,hi);	
	fput(pr_out3,n,stdout);	
	printf("%c%c",0x0d,0x0a);
#endif	
#ifdef HP    
	int n;

/* line 569 in the file print.c to: "n=(int)(width*scale*3+7)/8;" */
	n=(int)(width*scale*3+7)/8;
#ifdef CII
	printf("%c*b%dW",0x1b,n);
	fput(hp_out1,n,stdout);
	printf("%c*b%dW",0x1b,n);
	fput(hp_out2,n,stdout);
	printf("%c*b%dW",0x1b,n);
	fput(hp_out3,n,stdout);
#else
   printf("%c*b%dW",0x1b,n);
   fwrite(hp_out1,n,1,stdout);
   printf("%c*b%dW",0x1b,n);
   fwrite(hp_out2,n,1,stdout);
   printf("%c*b%dW",0x1b,n);
   fwrite(hp_out3,n,1,stdout);
   fflush(stdout);
#endif
#endif
}

output_line()	/* does buffering */
{
	static int n=0;

#ifdef PR
	if(!mono) {
		dither_image(7-n);
		n++;
		if(lowres)
			n++;
		if(n==8) {
			print_line();
			if(lowres)
				printf("%cJ%c",0x1b,0x11);	/* skip 17/216th of an inch */
			else
				printf("%cJ%c",0x1b,0x14);	/* skip 20/216th of an inch */
			n=0;
			memset(pr_out1,0,PRINTSIZE);
			memset(pr_out2,0,PRINTSIZE);
			memset(pr_out3,0,PRINTSIZE);
			}
		}
	else {
		mono_image(23-n);
		n++;
		if(lowres)
			n++;
		if(n==24) {
			print_line();
			if(lowres)
				printf("%cJ%c",0x1b,0x11);	/* skip 17/216th of an inch */
			else
				printf("%cJ%c",0x1b,0x14);	/* skip 20/216th of an inch */
			n=0;
			memset(pr_out1,0,PRINTSIZE);
			memset(pr_out2,0,PRINTSIZE);
			memset(pr_out3,0,PRINTSIZE);
			}
		}
#endif		
#ifdef HP
	if(!mono) {
		dither_image(n);
		print_line();
		memset(hp_out1,0,PRINTSIZE);
		memset(hp_out2,0,PRINTSIZE);
		memset(hp_out3,0,PRINTSIZE);
        }
	else {
		mono_image(n);
		n++;
		if(n==3) {
			print_line();
			memset(hp_out1,0,PRINTSIZE);
			memset(hp_out2,0,PRINTSIZE);
			memset(hp_out3,0,PRINTSIZE);
			}
		}
#endif

	return(n);
}
